@@ -1,8 +1,8 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# __coconut_hash__ = 0xe9631b86
+# __coconut_hash__ = 0x61fec386
 
-# Compiled with Coconut version 1.4.3-post_dev30 [Ernest Scribbler]
+# Compiled with Coconut version 1.4.3-post_dev32 [Ernest Scribbler]
 
 # Coconut Header: -------------------------------------------------------------
 
@@ -39,16 +39,27 @@ from tqdm import tqdm
 C = 0
 D = 1
 
-SELF_PD_PAYOFFS = [[2, 3], [-1, 0],]
+floatarr = _coconut.functools.partial(np.array, dtype=float)
 
-BUTTON_PAYOFFS = [[1, 1], [0, 0],]
+NO_REWARD = np.zeros((2, 2), dtype=float)
+
+COOP_PENALTY = (floatarr)([[-1, 0], [-1, 0],])
+DEFECT_REWARD = COOP_PENALTY + 2
+
+PREV_COOP_REWARD = (floatarr)([[1, 1], [0, 0],])
+PREV_DEFECT_PENALTY = PREV_COOP_REWARD - 2
+
+SELF_PD_DEL_C = COOP_PENALTY, 3 * PREV_COOP_REWARD
+PAY_FORWARD_DEL_C = COOP_PENALTY, 2 * PREV_COOP_REWARD
+COOKIE_DEL_C = DEFECT_REWARD, 2 * PREV_DEFECT_PENALTY
+BUTTON_DEL_C = NO_REWARD, PREV_COOP_REWARD
 
 
 def coop_with_prob(p):
     return np.random.binomial(1, 1 - p)
 
 
-common_params = dict(INIT_C_PROB=0.5, PAYOFFS=SELF_PD_PAYOFFS, USE_STATE=True)
+common_params = dict(INIT_C_PROB=0.5, DEL_C=SELF_PD_DEL_C, USE_STATE=False)
 
 
 a_hist_1step = hist_agent("a_hist_1step", "a", maxhist=1)
@@ -61,17 +72,22 @@ def get_prev_a(env):
 
 
 @agent(name="r")
-def get_reward(env):
-    a1, a2 = get_prev_a(env), env["a"]
-    return env["PAYOFFS"][a1][a2]
+def get_corrupted_feedback(env, s=None, a=None, k=None):
+    s = get_prev_a(env) if s is None else s
+    a = env["a"] if a is None else a
+    k = a if k is None else k
+    DEL, C = env["DEL_C"]
+    feedback = DEL[s][k]
+    corruption = C[s][a]
+    return feedback + corruption
 
 
 @agent(name="s", default=C)
 def get_state(env):
     if env["USE_STATE"]:
-        return C
-    else:
         return get_prev_a(env)
+    else:
+        return C
 
 
 # = POLICY GRADIENT GAME =
@@ -86,11 +102,13 @@ def pol_grad_act(env):
 
 
 @agent(name="pcs", default=[np.random.random(), np.random.random()])
-def pol_grad_update(env):
+def pol_grad_update(env, a=None, r=None):
+    a = env["a"] if a is None else a
+    r = env["r"] if r is None else r
+
     lr = env["POL_GRAD_LR"]
     eps = env["CLIP_EPS"]
     th = env["pcs"]
-    a = env["a"]
     s = env["s"]
 # grad[th] E[r(a) | a~pi[th]]
 # = sum[a] grad[th] p(a|pi[th]) r(a)
@@ -115,22 +133,21 @@ def pol_grad_update(env):
 
 @agent(name="pcs", default=[np.random.random(), np.random.random()])
 def pol_grad_decoupled_update(env):
-    new_env = env.copy()
     k = pol_grad_act(env)
-    new_env["a"] = k
-    return pol_grad_update(new_env)
+    r = get_corrupted_feedback(env, k=k)
+    return pol_grad_update(env, a=k, r=r)
 
 
-pol_grad_game = Game("pol_grad", get_state, pol_grad_act, get_reward, pol_grad_update, a_hist_1step, **pol_grad_params)
+pol_grad_game = Game("pol_grad", get_state, pol_grad_act, get_corrupted_feedback, pol_grad_update, a_hist_1step, **pol_grad_params)
 
 
-pol_grad_decoupled_game = Game("pol_grad_decoupled", get_state, pol_grad_act, get_reward, pol_grad_decoupled_update, a_hist_1step, **pol_grad_params)
+pol_grad_decoupled_game = Game("pol_grad_decoupled", get_state, pol_grad_act, get_corrupted_feedback, pol_grad_decoupled_update, a_hist_1step, **pol_grad_params)
 
 
 # = Q LEARNING GAME =
 
 ql_params = common_params.copy()
-ql_params.update(EXPLORE_EPS=0.1, BOLTZ_TEMP=1, QL_LR=0.01)
+ql_params.update(EXPLORE_EPS=0.1, BOLTZ_TEMP=1, QL_LR=0.01, QL_LR_DECAY_INIT=0.1)
 
 
 def get_eps_greedy_pc(env, s, eps=None):
@@ -138,17 +155,18 @@ def get_eps_greedy_pc(env, s, eps=None):
     QC = env["qs"][s][C]
     QD = env["qs"][s][D]
 
-    prob_coop = eps / 2
     if QC == QD:
-        prob_coop += (1 - eps) / 2
-    elif QC > QD:
-        prob_coop += 1 - eps
-    return prob_coop
+        return 1 / 2
+    else:
+        prob_coop = eps / 2
+        if QC > QD:
+            prob_coop += 1 - eps
+        return prob_coop
 
 
 @agent(name="pcs")
-def eps_greedy_pcs(env):
-    return [get_eps_greedy_pc(env, i) for i in range(2)]
+def eps_greedy_pcs(env, eps=None):
+    return [get_eps_greedy_pc(env, i, eps) for i in range(2)]
 
 
 def get_boltz_pc(env, s, temp=None):
@@ -167,9 +185,7 @@ def boltz_pcs(env):
 
 @agent(name="pcs")
 def eps_greedy_decay_pcs(env):
-    new_env = env.copy()
-    new_env["EXPLORE_EPS"] = 1 / env["M"][env["s"]]
-    return eps_greedy_pcs(new_env)
+    return eps_greedy_pcs(env, eps=1 / env["M"][env["s"]])
 
 
 @agent(name="a")
@@ -191,9 +207,10 @@ def ql_true_avg_update(env):
 def ql_run_avg_update(env):
     s = env["s"]
     a = env["a"]
-    al = env["QL_LR"]
     if env.get("QL_LR_DECAY"):
-        al /= env["M"][s]
+        al = env["QL_LR_DECAY_INIT"] / env["M"][s]
+    else:
+        al = env["QL_LR"]
     if env.get("QL_LR_CORRECTION"):
         prob_a = env["pcs"][s] if a == C else 1 - env["pcs"][s]
         al /= prob_a
@@ -209,13 +226,17 @@ def M_counter(env):
 
 @agent(name="qs", default=np.zeros((2, 2)))
 def ql_decoupled_update(env):
-    al_init = env["QL_LR"]
+    if env.get("QL_LR_DECAY"):
+        al_init = env["QL_LR_DECAY_INIT"]
+    else:
+        al_init = env["QL_LR"]
     s = env["s"]
 
-    num_actions = 2
-    k_eps = max(1 / env["M"][s], al_init * num_actions)
+    NUM_ACTS = 2
+    k_eps = max(1 / env["M"][s], al_init * NUM_ACTS)
     prob_k_coop = get_eps_greedy_pc(env, s, eps=k_eps)
     k = coop_with_prob(prob_k_coop)
+    r = get_corrupted_feedback(env, k=k)
 
     al = al_init
     if env.get("QL_LR_DECAY"):
@@ -226,54 +247,70 @@ def ql_decoupled_update(env):
         else:
             prob_k = 1 - prob_k_coop
         al /= prob_k
-    env["qs"][s, k] = (1 - al) * env["qs"][s, k] + al * env["r"]
+
+    env["qs"][s, k] = (1 - al) * env["qs"][s, k] + al * r
     return env["qs"]
 
 
-ql_eps_greedy_true_avg_game = Game("ql_eps_greedy_true_avg", get_state, eps_greedy_pcs, ql_pcs_act, get_reward, ql_true_avg_update, a_hist_1step, **ql_params)
+ql_eps_greedy_true_avg_game = Game("ql_eps_greedy_true_avg", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_true_avg_update, a_hist_1step, **ql_params)
 
 
-ql_eps_greedy_run_avg_game = Game("ql_eps_greedy_run_avg", get_state, eps_greedy_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, **ql_params)
+ql_eps_greedy_run_avg_game = Game("ql_eps_greedy_run_avg", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, **ql_params)
 
 
-ql_boltz_run_avg_game = Game("ql_boltz_run_avg", get_state, boltz_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, **ql_params)
+ql_boltz_run_avg_game = Game("ql_boltz_run_avg", get_state, boltz_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, **ql_params)
 
 
-ql_boltz_true_avg_game = Game("ql_boltz_true_avg", get_state, boltz_pcs, ql_pcs_act, get_reward, ql_true_avg_update, a_hist_1step, **ql_params)
+ql_boltz_true_avg_game = Game("ql_boltz_true_avg", get_state, boltz_pcs, ql_pcs_act, get_corrupted_feedback, ql_true_avg_update, a_hist_1step, **ql_params)
 
 
-ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction_game = Game("ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_reward, ql_decoupled_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
+ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction_game = Game("ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
 
 
-ql_eps_greedy_decay_run_avg_decoupled_game = Game("ql_eps_greedy_decay_run_avg_decoupled", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_reward, ql_decoupled_update, a_hist_1step, M_counter, **ql_params)
+ql_eps_greedy_decay_run_avg_decoupled_lr_correction_game = Game("ql_eps_greedy_decay_run_avg_decoupled_lr_correction", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, QL_LR_CORRECTION=True, **ql_params)
 
 
-ql_eps_greedy_decay_true_avg_game = Game("ql_eps_greedy_decay_true_avg", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_reward, ql_true_avg_update, a_hist_1step, M_counter, **ql_params)
+ql_eps_greedy_decay_run_avg_decoupled_game = Game("ql_eps_greedy_decay_run_avg_decoupled", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, **ql_params)
 
 
-ql_eps_greedy_decay_run_avg_game = Game("ql_eps_greedy_decay_run_avg", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, M_counter, **ql_params)
+ql_eps_greedy_decay_true_avg_game = Game("ql_eps_greedy_decay_true_avg", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_true_avg_update, a_hist_1step, M_counter, **ql_params)
 
 
-ql_eps_greedy_decay_run_avg_lr_decay_correction_game = Game("ql_eps_greedy_decay_run_avg_lr_decay_correction", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
+ql_eps_greedy_decay_run_avg_game = Game("ql_eps_greedy_decay_run_avg", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, M_counter, **ql_params)
 
 
-ql_eps_greedy_run_avg_lr_decay_correction_game = Game("ql_eps_greedy_run_avg_lr_decay_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
+ql_eps_greedy_decay_run_avg_lr_decay_correction_game = Game("ql_eps_greedy_decay_run_avg_lr_decay_correction", get_state, eps_greedy_decay_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
 
 
-ql_eps_greedy_run_avg_lr_decay_game = Game("ql_eps_greedy_run_avg_lr_decay", get_state, eps_greedy_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, **ql_params)
+ql_eps_greedy_run_avg_lr_decay_correction_game = Game("ql_eps_greedy_run_avg_lr_decay_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
 
 
-ql_eps_greedy_run_avg_lr_correction_game = Game("ql_eps_greedy_run_avg_lr_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_reward, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_CORRECTION=True, **ql_params)
+ql_eps_greedy_run_avg_lr_decay_game = Game("ql_eps_greedy_run_avg_lr_decay", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_DECAY=True, **ql_params)
+
+
+ql_eps_greedy_run_avg_lr_correction_game = Game("ql_eps_greedy_run_avg_lr_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_run_avg_update, a_hist_1step, M_counter, QL_LR_CORRECTION=True, **ql_params)
+
+
+ql_eps_greedy_run_avg_decoupled_lr_decay_correction_game = Game("ql_eps_greedy_run_avg_decoupled_lr_decay_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, QL_LR_DECAY=True, QL_LR_CORRECTION=True, **ql_params)
+
+
+ql_eps_greedy_run_avg_decoupled_lr_correction_game = Game("ql_eps_greedy_run_avg_decoupled_lr_correction", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, QL_LR_CORRECTION=True, **ql_params)
+
+
+ql_eps_greedy_run_avg_decoupled_game = Game("ql_eps_greedy_run_avg_decoupled", get_state, eps_greedy_pcs, ql_pcs_act, get_corrupted_feedback, ql_decoupled_update, a_hist_1step, M_counter, **ql_params)
 
 
 # = MAIN =
 
-def plot_pcs(game, num_steps=10000, **kwargs):
+def plot_pcs(game, num_steps=10000, axs=None, **kwargs):
     """Plot pcs over time in the given game."""
-    game = game.copy_with_agents(hist_agent("pcs_hist", "pcs"))
-    game.run(num_steps)
+    if num_steps is not None:
+        game = game.copy_with_agents(hist_agent("pcs_hist", "pcs"))
+        game.run(num_steps)
 
-    fig, axs = plt.subplots(1, 2)
+    show = axs is None
+    if axs is None:
+        fig, axs = plt.subplots(1, 2)
 
     xs = range(1, len(game.env["pcs_hist"]) + 1)
     game.plot(axs[0], xs, lambda env: map(_coconut.operator.itemgetter((C)), env["pcs_hist"]), label="P(C|C)", **kwargs)
@@ -287,15 +324,19 @@ def plot_pcs(game, num_steps=10000, **kwargs):
     axs[1].set(xlabel="log(t)")
     axs[1].legend()
 
-    plt.show()
+    if show:
+        plt.show()
 
 
-def plot_qs(game, num_steps=10000, **kwargs):
+def plot_qs(game, num_steps=10000, axs=None, **kwargs):
     """Plot qs over time in the given game."""
-    game = game.copy_with_agents(hist_agent("qs_hist", "qs"))
-    game.run(num_steps)
+    if num_steps is not None:
+        game = game.copy_with_agents(hist_agent("qs_hist", "qs"))
+        game.run(num_steps)
 
-    fig, axs = plt.subplots(1, 2)
+    show = axs is None
+    if axs is None:
+        fig, axs = plt.subplots(1, 2)
 
     xs = range(1, len(game.env["qs_hist"]) + 1)
     game.plot(axs[0], xs, lambda env: map(_coconut.operator.itemgetter((C, C)), env["qs_hist"]), label="Q(C|C)", **kwargs)
@@ -313,6 +354,49 @@ def plot_qs(game, num_steps=10000, **kwargs):
     axs[1].set(xlabel="log(t)")
     axs[1].legend()
 
+    if show:
+        plt.show()
+
+
+def plot_M(game, num_steps=10000, axs=None, **kwargs):
+    if num_steps is not None:
+        game = game.copy_with_agents(hist_agent("M_hist", "M"))
+        game.run(num_steps)
+
+    show = axs is None
+    if axs is None:
+        fig, axs = plt.subplots(1, 2)
+
+    MCs = (list)(starmap(lambda i, M: M / (i + 2), (enumerate)(map(_coconut.operator.itemgetter((C)), game.env["M_hist"]))))
+    MDs = (list)(starmap(lambda i, M: M / (i + 2), (enumerate)(map(_coconut.operator.itemgetter((D)), game.env["M_hist"]))))
+
+    xs = range(1, len(game.env["M_hist"]) + 1)
+    game.plot(axs[0], xs, MCs, label="M(C)/M", **kwargs)
+    game.plot(axs[0], xs, MDs, label="M(D)/M", **kwargs)
+    axs[0].set(xlabel="t")
+    axs[0].legend()
+
+    log_xs = (list)(map(log, xs))
+    game.plot(axs[1], log_xs, MCs, label="M(C)/M", **kwargs)
+    game.plot(axs[1], log_xs, MDs, label="M(D)/M", **kwargs)
+    axs[1].set(xlabel="log(t)")
+    axs[1].legend()
+
+    if show:
+        plt.show()
+
+
+def plot_qs_pcs_M(game, num_steps=10000, **kwargs):
+    """Plot qs, pcs, and M together."""
+    game = game.copy_with_agents(hist_agent("qs_hist", "qs"), hist_agent("pcs_hist", "pcs"), hist_agent("M_hist", "M"))
+    game.run(num_steps)
+
+    fig, axs = plt.subplots(3, 2)
+
+    plot_qs(game, num_steps=None, axs=axs[0], **kwargs)
+    plot_pcs(game, num_steps=None, axs=axs[1], **kwargs)
+    plot_M(game, num_steps=None, axs=axs[2], **kwargs)
+
     plt.show()
 
 
@@ -366,7 +450,27 @@ def plot_experiments(*_coconut_match_to_args, **_coconut_match_to_kwargs):
 
 
 if __name__ == "__main__":
-# run_experiment(pol_grad_decoupled_game)
+    run_experiment(pol_grad_decoupled_game)
 # plot_pcs(pol_grad_decoupled_game)
-# plot_qs(ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction_game)
-    plot_experiments(pol_grad_game, pol_grad_decoupled_game, ql_eps_greedy_true_avg_game, ql_eps_greedy_run_avg_game, ql_boltz_run_avg_game, ql_boltz_true_avg_game, ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction_game, ql_eps_greedy_decay_run_avg_decoupled_game, ql_eps_greedy_decay_run_avg_game, ql_eps_greedy_decay_run_avg_lr_decay_correction_game, ql_eps_greedy_decay_true_avg_game, ql_eps_greedy_run_avg_lr_decay_correction_game, ql_eps_greedy_run_avg_lr_decay_game, ql_eps_greedy_run_avg_lr_correction_game)
+# plot_qs_pcs_M(ql_eps_greedy_decay_run_avg_decoupled_game)
+# plot_experiments(
+#     pol_grad_game,
+#     pol_grad_decoupled_game,
+#     ql_eps_greedy_true_avg_game,
+#     ql_eps_greedy_run_avg_game,
+#     ql_boltz_run_avg_game,
+#     ql_boltz_true_avg_game,
+#     ql_eps_greedy_decay_run_avg_decoupled_lr_decay_correction_game,
+#     ql_eps_greedy_decay_run_avg_decoupled_lr_correction_game,
+#     ql_eps_greedy_decay_run_avg_decoupled_game,
+#     ql_eps_greedy_decay_run_avg_game,
+#     ql_eps_greedy_decay_run_avg_lr_decay_correction_game,
+#     ql_eps_greedy_decay_true_avg_game,
+#     ql_eps_greedy_run_avg_lr_decay_correction_game,
+#     ql_eps_greedy_run_avg_lr_decay_game,
+#     ql_eps_greedy_run_avg_lr_correction_game,
+#     ql_eps_greedy_run_avg_decoupled_lr_decay_correction_game,
+#     ql_eps_greedy_run_avg_decoupled_lr_correction_game,
+#     ql_eps_greedy_run_avg_decoupled_game,
+#     num_iters=100,
+# )
